/*
 * Copyright 2015 Higher Frequency Trading
 *
 * http://www.higherfrequencytrading.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.openhft.chronicle.threads;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Marker interface to show we support pause(long timeout, TimeUnit timeUnit)
 */
public interface TimingPauser extends Pauser {

    /**
     * Pauses but keep tracks of accumulated pause time and throws if timeout exceeded
     * @param timeout timeout
     * @param timeUnit unit
     * @throws TimeoutException thrown if timeout passes
     */
    void pause(long timeout, TimeUnit timeUnit) throws TimeoutException;
}
